{% extends "!page.html" %}
{% block body %}
{% if current_version and latest_version and current_version != latest_version %}
<div class="admonition warning">
<!-- <p class="admonition-title">Warning</p> -->
  <p>
    {% if current_version.is_released %}
    You're reading <strong>an old version</strong> of this documentation.
    For up-to-date information, please have a look at <a href="{{ vpathto(latest_version.name) }}">{{latest_version.name}}</a>.
    {% else %}
    You're reading the documentation for a <strong>development version</strong>.
    For the latest released version, please have a look at <a href="{{ vpathto(latest_version.name) }}">{{latest_version.name}}</a>.
    {% endif %}
  </p>
</div>
{% endif %}
{{ super() }}
{% endblock %}